# coding: utf-8
"""

    ファイル名：   BridgedGuinierPorod.py

    Copyright (c) 2017, Masatsuyo Takahashi, KEK-PF

"""

import os
import re
import time
import numpy                as np
from scipy                  import optimize
from bisect                 import bisect_right
import matplotlib.pyplot    as plt
import seaborn              as sns
from lmfit                  import minimize, Parameters, Parameter, report_fit
from IntensityData          import IntensityData
from BridgeFunctions        import Catenary     as Brige
from GuinierPorodFit        import GuinierPorodFit
from SimulatedAnnealing     import SimulatedAnnealing

DEBUG = False
DEBUG_PLOT = False

def debug_plot_guinier( x, y, gx, gy ):
    plt.plot( x, y )
    plt.plot( gx, gy )
    plt.show( block=True )
    time.sleep( 0.5 )
    plt.close()

def debug_plot_porod( x, y, px, py ):
    plt.plot( x, y )
    plt.plot( px, py )
    plt.show( block=True  )
    time.sleep( 0.5 )
    plt.close()

def debug_plot_brige( x, y, y_, bridge ):
    plt.plot( x, y )
    plt.plot( x, y_ )
    p1  = bridge.p1
    p2  = bridge.p2
    plt.plot( [ x[p1], x[p2] ], [ y_[p1], y_[p2] ], 'o', color='red' )
    plt.show( block=True  )
    time.sleep( 0.5 )
    plt.close()

class Guinier:
    def __init__( self, x, y, w ):
        self.x  = x
        self.y  = y
        self.w  = w

    def fit( self ):
        params = Parameters()
        G_init  = np.average( self.y[5:10] )
        params.add('G',     value= G_init,      min=1e-5,   max=G_init*100 )
        params.add('Rg',    value= 100,         min=1.0,    max=500.0 )

        def func_Guinier( p, x, y, w ):
            G   = p['G'].value
            Rg  = p['Rg'].value
            # print( 'func_Guinier: G, Rg=', G, Rg )

            y_ = G * np.exp( -x**2 * Rg**2/3 )
            return w*( y_ - y )

        result = minimize( func_Guinier, params, args=( self.x, self.y, self.w ) )
        self.G  = result.params['G'].value
        self.Rg = result.params['Rg'].value
        print( 'G, Rg=', self.G, self.Rg )

    def __call__( self, x ):
        return self.G * np.exp( -x**2 * self.Rg**2/3 )

class Porod:
    def __init__( self, x, y, w ):
        self.x  = x
        self.y  = y
        self.w  = w

    def fit( self ):
        params = Parameters()
        D_init  = np.average( self.y[0:5] ) * self.x[0] ** 2
        params.add('D',     value= D_init*1e-5,     min=1e-5,      max=D_init*1e5 )
        params.add('d',     value= 3.0,             min=1.0,    max=5.0 )

        def func_Porod( p, x, y, w ):
            D   = p['D'].value
            d   = p['d'].value
            # print( 'func_Porod: D, d=', D, d )

            y_ = D / np.power( x, d )
            return w*( y_ - y )

        result = minimize( func_Porod, params, args=( self.x, self.y, self.w ) )
        self.D  = result.params['D'].value
        self.d  = result.params['d'].value
        print( 'D, d=', self.D, self.d )

    def __call__( self, x ):
        return self.D / np.power( x, self.d )

def _BridgedGuinierPorod( bridge, G, Rg, D, d, x, return_range=False, fast=False ):

    q1, q2 = bridge.q1, bridge.q2
    assert( q1 < q2 )

    q1_q2 = np.logical_and( x >= q1, x < q2  )

    y_  = np.hstack( [  G * np.exp( - x[ x < q1 ]**2 * Rg**2 / 3 ),     # Guinier
                        bridge.predict( x[ q1_q2 ], fast=fast ),        # Bridge
                        D / np.power( x[ x >= q2 ], d )                 # Porod
                     ] )

    if return_range:
        return y_, q1_q2
    else:
        return y_

class BridgedGuinierPorod:
    def __init__( self, intensity ):
        self.intensity = intensity

        self.x = x = self.intensity.Q
        self.y = y = self.intensity.I
        # I_ = intensity.smoother.predict( self.intensity.Q )
        # y = I_
        e = self.intensity.E
        self.w = w = 1/e**2

        self.bridge = Brige( intensity.Q, intensity.I )

    def get_params( self, q1, q2 ):
        p1  = self.bridge.get_point_index( q1 )
        slice_ = slice(0, p1)

        gx  = self.x[slice_]
        gunier = Guinier( gx, self.y[slice_], self.w[slice_] )
        gunier.fit()
        gy_ = gunier( gx )

        if DEBUG_PLOT:
            debug_plot_guinier( self.x, self.y, gx, gy_ )

        G   = gunier.G
        Rg  = gunier.Rg

        p2  = self.bridge.get_point_index( q2 )
        slice_ = slice( p2, None )

        px  = self.x[slice_]
        porod = Porod( px, self.y[slice_], self.w[slice_] )
        porod.fit()
        py_ = porod( px )

        if DEBUG_PLOT:
            debug_plot_porod( self.x, self.y, px, py_ )

        D   = porod.D
        d   = porod.d

        return G, Rg, D, d

    def get_params_fast( self, q1, q2 ):
        p1  = self.bridge.get_point_index( q1 )
        Yg  = np.log( [ self.y[0], self.y[p1] ] )
        Xg  = np.array( [ self.x[0], self.x[p1] ] )**2
        slope_g      = ( Yg[1] - Yg[0] ) / ( Xg[1] - Xg[0] )    # in Guinier Plot
        intercept_g  = Yg[0] - slope_g * Xg[0]
        Rg  = np.sqrt( -3 * slope_g )
        G   = np.exp( intercept_g )

        p2  = self.bridge.get_point_index( q2 )
        Yp  = np.log( [ self.y[p2], self.y[-1] ] )
        Xp  = np.log( [ self.x[p2], self.x[-1] ] )
        slope_p      = ( Yp[1] - Yp[0] ) / ( Xp[1] - Xp[0] )    # in Porod Plot
        intercept_p  = Yp[0] - slope_p * Xp[0]
        d   = -slope_p
        D   = np.exp( intercept_p )

        return G, Rg, D, d

    def evaluate_diff( self, q1, q2, anneal=False, fast=False ):
        if q1 < q2:
            if anneal:
                fast_ = False
                G, Rg, D, d = self.get_params_fast( q1, q2 )
                self.bridge.fit( q1, q2, params=( G, Rg, D, d ), fast=fast_ )
            else:
                G, Rg, D, d = self.get_params( q1, q2 )
                # print( '(1) G, Rg, D, d=', G, Rg, D, d )
                self.bridge.fit( q1, q2, params=( G, Rg, D, d ) )
                fast_ = False
            y_  = _BridgedGuinierPorod( self.bridge, G, Rg, D, d, self.x, fast=fast_ )

            if DEBUG_PLOT:
                debug_plot_brige( self.x, y, self.y_, self.bridge )

            return ( y_ - self.y )
        else:
            return np.ones( y.shape[0] ) * 1e100

    def func_Whole( self, p2 ):
        q1  = p2['q1'].value
        q2  = p2['q2'].value
        print( 'func: q1, q2=', q1, q2 )
        return self.evaluate_diff( q1, q2 )

    def fit( self, anneal=False ):
        if anneal:
            self.fit_anneal()
        else:
            params = Parameters()
            params.add('q1',    value= self.x[20],   min=self.x[10], max=self.x[-1]/2 )
            params.add('q2',    value= self.x[140],  min=self.x[50], max=self.x[-1] )

            result = minimize( self.func_Whole, params, args=() )

            self.q1 = result.params['q1'].value
            self.q2 = result.params['q2'].value

        G, Rg, D, d = self.get_params( self.q1, self.q2 )
        print( '(2) G, Rg, D, d=', G, Rg, D, d )

        self.G  = G
        self.Rg = Rg
        self.D  = D
        self.d  = d

        self.bridge.fit( self.q1, self.q2, params=( G, Rg, D, d ) )
        self.p1 = self.bridge.p1
        self.p2 = self.bridge.p2

        # self.G  = self.bridge.predict( self.q1 ) / np.exp( -self.q1**2*self.Rg**2/3 )

        print( 'Rg, d=', self.Rg, self.d )
        print( 'q1, q2=', self.q1, self.q2 )

    def anneal_objective( self, q ):
        q1, q2 = q
        # print( 'anneal_objective: q1, q2=', q1, q2 )
        diff = self.evaluate_diff( q1, q2, anneal=True, fast=True )
        return np.log( np.sum( diff**2 ) )
        # return np.log( np.sum( ( self.intensity.W* diff) **2 ) )

    def fit_anneal( self ):

        t_start = time.time()

        def xconstaints( x ):
            return ( x[0] + 0.01 < x[1] ) and ( x[1] - x[0] < 0.3 )

        anneal = SimulatedAnnealing()
        anneal.minimize( self.anneal_objective, xrange=[ [ self.x[5], self.x[-1]/8 ], [ self.x[20], self.x[-5] ] ],
                            start=[ 0.01, 0.1 ], seed=1234,
                            n=50, m=10,
                            xconstaints=xconstaints )

        t_end = time.time()

        print( 'anneal.xc=', anneal.xc, t_end - t_start  )
        self.q1 = anneal.xc[0]
        self.q2 = anneal.xc[1]

    def predict( self, x ):
        return _BridgedGuinierPorod( self.bridge, self.G, self.Rg, self.D, self.d, x )

def demo_plot_compare( file, anneal=False ):

    intensity = IntensityData( file, add_smoother=True )
    d, f = os.path.split( file )

    Y_ = np.log( intensity.smoother( intensity.Q ) )

    crysol_log_re = re.compile( r'Shape\s+Rg\s+:\s+(\d+\.\d*)' )

    if intensity.comments is None:
        crysol_Rg = None
    else:
        m = crysol_log_re.search( intensity.comments )
        if m:
            crysol_Rg = m.group(1)
            print( 'Shape Rg=', crysol_Rg )

    model = BridgedGuinierPorod( intensity )
    if anneal:
        model.fit( anneal=True )
    else:
        model.fit()

    print( 'model: q1, q2=', model.q1, model.q2 )

    gpfit = GuinierPorodFit( intensity )
    gpfit.fit()

    fig = plt.figure( figsize=(16,6) )
    ax1 = fig.add_subplot( 121 )
    ax2 = fig.add_subplot( 122 )

    if crysol_Rg is None:
        clabel  = 'Experiment Data'
    else:
        clabel  = 'CRYSOL Generated Data ( Rg=%s )' % ( crysol_Rg )
    slabel  = 'Smoothed Curve'
    blabel  = 'Bridged Guinier-Porod fit ( Rg=%.4g, d=%.3g )' % ( model.Rg, model.d )
    gplabel = 'Guinier-Porod fit ( Rg=%.4g, d=%.3g )' % ( gpfit.Rg, gpfit.d )

    for i, ax in enumerate( [ ax1, ax2 ] ):

        ax.set_title( 'Bridged Guinier-Porod demo for ' + f )

        ax.plot( intensity.X, intensity.Y, 'o', markersize=4, label=clabel )
        ax.plot( intensity.X, Y_, color='yellow', label=slabel )

        ax.set_xlabel( 'Q**2' )
        ax.set_ylabel( 'Ln(I)' )

        MY =  np.log( model.predict( intensity.Q ) )
        # MY =  np.log( _BridgedGuinierPorod( model.bridge, model.G, model.Rg, model.D, model.d, intensity.Q ) )
        ax.plot( intensity.X, MY, color='cyan', label=blabel )

        ax.plot( intensity.X, np.log( gpfit.model( intensity.Q ) ), color='pink', label=gplabel )
        p1 = model.p1
        p2 = model.p2
        bx = [ intensity.X[p1], intensity.X[p2] ]
        by = [ MY[p1], MY[p2] ]
        ax.plot( bx, by, 'o', color='red' )
        ax.plot( intensity.X[p1:p2], MY[p1:p2], color='red', label='Bridge Curve' )
        ax.legend()

    xmin, xmax = ax1.get_xlim()
    ymin, ymax = ax1.get_ylim()

    Ymax = np.max( intensity.Y )

    xmin2 = -0.0005
    xmax2 = xmax * 1/32
    ymin2 = ymin * 1/4 + ymax * 3/4 - 3
    ymax2 = Ymax + ( ymax - ymin ) *0.02

    ax2.set_xlim( xmin2, xmax2 )
    ax2.set_ylim( ymin2, ymax2)

    plt.tight_layout()
    plt.show()
