# coding: utf-8
"""

    ファイル名：   BridgedGuinierPorod.py

    Copyright (c) 2017, Masatsuyo Takahashi, KEK-PF

"""

import os
import re
import time
import warnings
import numpy                as np
from scipy                  import stats, optimize
from bisect                 import bisect_right
import matplotlib.pyplot    as plt
import seaborn              as sns
from lmfit                  import minimize, Parameters, Parameter, report_fit
from IntensityData          import IntensityData
from BridgeCurve            import BridgeCurve
from GuinierPorodFit        import GuinierPorodFit

warnings.filterwarnings("ignore")

DEBUG = False
DEBUG_PLOT = False
Q1_PENALITY_LIMIT   = 0.005
Q1_PENALITY_SCALE   = 1e-6
MIN_BRIDE_LENGTH    = 0.02
MAX_BRIDE_LENGTH    = 0.4

def debug_plot_guinier( x, y, gx, gy ):
    plt.plot( x, y )
    plt.plot( gx, gy )
    plt.show( block=False )
    time.sleep( 0.5 )
    plt.close()

def debug_plot_porod( x, y, px, py ):
    plt.plot( x, y )
    plt.plot( px, py )
    plt.show( block=False  )
    time.sleep( 0.5 )
    plt.close()

def debug_plot_brige( x, y, y_, bridge ):
    plt.plot( x, y )
    plt.plot( x, y_ )
    p1  = bridge.p1
    p2  = bridge.p2
    plt.plot( [ x[p1], x[p2] ], [ y_[p1], y_[p2] ], 'o', color='red' )
    plt.show( block=True  )
    time.sleep( 0.5 )
    plt.close()

class Guinier:
    def __init__( self, x, y, w ):
        self.x  = x
        self.y  = y
        self.w  = w

    def fit( self ):
        params = Parameters()
        G_init  = np.average( self.y[5:10] )
        params.add('G',     value= G_init,      min=1e-5,   max=G_init*100 )
        params.add('Rg',    value= 100,         min=1.0,    max=500.0 )

        def func_Guinier( p, x, y, w ):
            G   = p['G'].value
            Rg  = p['Rg'].value
            # print( 'func_Guinier: G, Rg=', G, Rg )

            y_ = G * np.exp( -x**2 * Rg**2/3 )
            return w*( y_ - y )

        result = minimize( func_Guinier, params, args=( self.x, self.y, self.w ) )
        self.G  = result.params['G'].value
        self.Rg = result.params['Rg'].value
        if DEBUG: print( 'G, Rg=', self.G, self.Rg )

    def __call__( self, x ):
        return self.G * np.exp( -x**2 * self.Rg**2/3 )

class Porod:
    def __init__( self, x, y, w ):
        self.x  = x
        self.y  = y
        self.w  = w

    def fit( self ):
        params = Parameters()
        d_init = 3.7
        # d_init = 4
        D_init  = np.average( self.y[0:3] ) * self.x[0] ** d_init
        params.add('D',     value= D_init,      min=D_init*1e-3,   max=D_init*1e3 )
        params.add('d',     value= d_init,      min=1.0,    max=5.0 )

        def func_Porod( p, x, y, w ):
            D   = p['D'].value
            d   = p['d'].value
            # print( 'func_Porod: D, d=', D, d )

            y_ = D / np.power( x, d )
            return w*( y_ - y )

        result = minimize( func_Porod, params, args=( self.x, self.y, self.w ) )
        self.D  = result.params['D'].value
        self.d  = result.params['d'].value
        if DEBUG:
            print( 'Prod: D, d=', self.D, self.d, 'D/np.power(q2, d)=' )
            for i in [ 0, -1 ]:
                Py = self.D/np.power( self.x[i], self.d )
                print( 'Porod: ', i,  Py, self.y[i], Py/self.y[i] )

    def __call__( self, x ):
        return self.D / np.power( x, self.d )

def _BridgedGuinierPorod( bridge, G, Rg, D, d, x, return_range=False ):

    q1, q2 = bridge.q1, bridge.q2
    assert( q1 < q2 )

    q1_q2 = np.logical_and( x >= q1, x < q2  )
    by = bridge.predict( x[ q1_q2 ] )

    y_  = np.hstack( [  G * np.exp( - x[ x < q1 ]**2 * Rg**2 / 3 ),     # Guinier
                        by,                                             # Bridge
                        D / np.power( x[ x >= q2 ], d )                 # Porod
                     ] )

    if return_range:
        return y_, q1_q2
    else:
        return y_

LARGE_VALUE = 1e100

class BridgedGuinierPorod:
    def __init__( self, intensity ):
        self.intensity = intensity
        I_ = intensity.smoother.predict( self.intensity.Q )

        self.bridge = BridgeCurve( intensity.Q, intensity.I, smoother=intensity.smoother )
        self.min_bridge_width = ( self.intensity.Q[1] - self.intensity.Q[0] )*10
        self.q_max = intensity.Q[-1]
        print( 'min_bridge_width=', self.min_bridge_width )

    def get_params( self, q1, q2 ):
        p1  = self.bridge.get_point_index( q1 )
        slice_ = slice(0, p1)

        gx  = self.x[slice_]
        gunier = Guinier( gx, self.y[slice_], self.w[slice_] )
        gunier.fit()
        gy_ = gunier( gx )

        if DEBUG_PLOT:
            debug_plot_guinier( self.x, self.y, gx, gy_ )

        G   = gunier.G
        Rg  = gunier.Rg

        p2  = self.bridge.get_point_index( q2 )
        slice_ = slice( p2, None )

        px  = self.x[slice_]
        porod = Porod( px, self.y[slice_], self.w[slice_] )
        porod.fit()
        py_ = porod( px )

        if DEBUG_PLOT:
            debug_plot_porod( self.x, self.y, px, py_ )

        D   = porod.D
        d   = porod.d

        return G, Rg, D, d

    def get_params_fast( self, q1, q2 ):
        p1  = self.bridge.get_point_index( q1 )
        Yg  = np.log( self.y[3:p1] )
        Xg  = self.x[3:p1]**2
        slope_g, intercept_g, _, _, _ = stats.linregress( Xg, Yg )
        Rg  = np.sqrt( -3 * slope_g )
        G   = np.exp( intercept_g )

        p2  = self.bridge.get_point_index( q2 )
        Yp  = np.log( self.y[p2:] )
        Xp  = np.log( self.x[p2:] )
        slope_p, intercept_p, _, _, _ = stats.linregress( Xp, Yp )
        d   = -slope_p
        D   = np.exp( intercept_p )
        return G, Rg, D, d

    def evaluate_diff( self, q1, q2 ):
        G, Rg, D, d = self.get_params( q1, q2 )
        # G, Rg, D, d = self.get_params_fast( q1, q2 )
        if np.isnan( Rg ) or q2 - q1 < self.min_bridge_width:
            return np.ones( self.y.shape[0] ) * LARGE_VALUE

        if DEBUG: print( '(1) G, Rg, D, d=', G, Rg, D, d )

        self.bridge.fit( q1, q2, params=( G, Rg, D, d ) )
        y_, q1_q2 = _BridgedGuinierPorod( self.bridge, G, Rg, D, d, self.x, return_range=True )
        weights = np.ones( len(y_) )
        # TODO: separate evaluation for Guinier, Bridge, Porod rerions
        weights[0:self.bridge.p1] *= 5
        # weights[q1_q2] *= 0.1

        scale = 1
        if q1 < Q1_PENALITY_LIMIT:
            slice_ = slice(0, self.bridge.p1)
            x = self.intensity.X[slice_]
            y = self.intensity.Y[slice_]
            _, _, r_value, _, _ = stats.linregress( x, y )
            # if it is closer to linearity, i.e. r_value is closer to -1,
            # then the penalty should be larger
            # because there is no need to narrow the Guinier region.
            penalty = 1/( 1 + r_value )/q1 * Q1_PENALITY_SCALE
            print( 'penalty=', penalty, 'q1=', q1 )
            scale += penalty

        if DEBUG_PLOT:
            debug_plot_brige( self.x, self.y, y_, self.bridge )

        return self.w * ( y_ - self.y ) * weights * scale

    def func_Whole( self, p2 ):
        q1  = p2['q1'].value
        q2  = p2['q2'].value
        if DEBUG: print( 'func: q1, q2=', q1, q2 )
        return self.evaluate_diff( q1, q2 )

    def fit( self, q1_init, q2_init, anneal=False ):
        self.x = x = self.intensity.Q
        self.y = y = self.intensity.I
        # y = I_
        e = self.intensity.E
        self.w = w = 1/e

        if anneal:
            self.fit_anneal( q1_init, q2_init )
        else:
            params = Parameters()
            params.add('q1',    value= q1_init,     min=x[10], max=x[-1]/2 )
            params.add('q2',    value= q2_init,     min=x[50], max=x[-1] )

            result = minimize( self.func_Whole, params, args=() )

            self.q1 = result.params['q1'].value
            self.q2 = result.params['q2'].value

        G, Rg, D, d = self.get_params( self.q1, self.q2 )
        if DEBUG: print( '(2) G, Rg, D, d=', G, Rg, D, d )

        self.G  = G
        self.Rg = Rg
        self.D  = D
        self.d  = d

        self.bridge.fit( self.q1, self.q2, params=( G, Rg, D, d ) )
        self.p1 = self.bridge.p1
        self.p2 = self.bridge.p2

        if DEBUG:
            print( 'Rg, d=', self.Rg, self.d )
            print( 'q1, q2=', self.q1, self.q2 )

    def anneal_objective( self, q ):
        q1, q2 = q
        # print( 'anneal_objective: q1, q2=', q1, q2 )
        diff = self.evaluate_diff( q1, q2 )
        return np.sum( diff**2 )

    def fit_anneal( self, q1_init, q2_init ):
        from SimulatedAnnealing     import SimulatedAnnealing

        t_start = time.time()

        def xconstaints( x ):
            w = x[1] - x[0]
            return ( w >= MIN_BRIDE_LENGTH ) and ( w <= MAX_BRIDE_LENGTH )

        anneal = SimulatedAnnealing()
        anneal.minimize( self.anneal_objective, xrange=[ [ self.x[5], self.x[-1]/8 ], [ self.x[20], self.x[-5] ] ],
                            start=[ q1_init, q2_init ], seed=1234,
                            n=50, m=50,
                            xconstaints=xconstaints )

        t_end = time.time()

        print( 'anneal.xc=', anneal.xc, t_end - t_start  )
        self.q1 = anneal.xc[0]
        self.q2 = anneal.xc[1]

    def predict( self, x ):
        return _BridgedGuinierPorod( self.bridge, self.G, self.Rg, self.D, self.d, x )

PLOT_SMOOTHED_CURVE = False

def demo_plot_compare( file, q1_init, q2_init, auto=False, anneal=False ):

    intensity = IntensityData( file, add_smoother=True  )
    d, f = os.path.split( file )

    Y_ = np.log( intensity.smoother( intensity.Q ) )

    crysol_log_re = re.compile( r'Shape\s+Rg\s+:\s+(\d+\.\d*)' )

    if intensity.comments is None:
        crysol_Rg = None
    else:
        m = crysol_log_re.search( intensity.comments )
        if m:
            crysol_Rg = m.group(1)
            print( 'Shape Rg=', crysol_Rg )
        else:
            crysol_Rg = None

    model = BridgedGuinierPorod( intensity )
    model.fit( q1_init, q2_init, anneal=anneal )

    print( 'model: q1, q2=', model.q1, model.q2 )
    print( 'model: Rg,  d=', model.Rg, model.d )
    print( 'model: e,  f=', model.bridge.e, model.bridge.f )

    gpfit = GuinierPorodFit( intensity )
    gpfit.fit()

    fig = plt.figure( figsize=(16,6) )
    ax1 = fig.add_subplot( 121 )
    ax2 = fig.add_subplot( 122 )

    if crysol_Rg is None:
        clabel  = 'Experiment Data'
    else:
        clabel  = 'CRYSOL Generated Data ( Rg=%s )' % ( crysol_Rg )
    slabel  = 'Smoothed Curve'
    blabel  = 'Bridged Guinier-Porod fit ( Rg=%.4g, d=%.3g )' % ( model.Rg, model.d )
    gplabel = 'Guinier-Porod fit ( Rg=%.4g, d=%.3g )' % ( gpfit.Rg, gpfit.d )

    for i, ax in enumerate( [ ax1, ax2 ] ):

        ax.set_title( 'Bridged Guinier-Porod demo for ' + f )
        ax.set_xlabel( 'Q**2' )
        ax.set_ylabel( 'Ln(I)' )

        ax.plot( intensity.X, intensity.Y, 'o', markersize=4, label=clabel )

        if PLOT_SMOOTHED_CURVE:
            ax.plot( intensity.X, Y_, color='yellow', label=slabel )

        ax.plot( intensity.X, np.log( gpfit.model( intensity.Q ) ), color='pink', label=gplabel )

        MY =  np.log( model.predict( intensity.Q ) )
        # MY =  np.log( _BridgedGuinierPorod( model.bridge, model.G, model.Rg, model.D, model.d, intensity.Q ) )
        ax.plot( intensity.X, MY, color='cyan', label=blabel )

        bx_ = np.linspace( model.q1, model.q2, 400 )
        bX_ = bx_**2
        bY_ = np.log( model.predict( bx_ ) )

        bx = [ bX_[0],  bX_[-1] ]
        by = [ bY_[0],  bY_[-1] ]
        ax.plot( bx, by, 'o', color='red' )
        ax.plot( bX_, bY_, color='red', label='Bridge Curve' )

        ax.legend()

    xmin, xmax = ax1.get_xlim()
    ymin, ymax = ax1.get_ylim()

    Ymax = np.max( intensity.Y )

    print( 'xmin, xmax=', xmin, xmax )

    xmin2 = -0.0005 if xmin < -0.01 else -0.0001
    xmax2 = xmax * 1/64
    ymin2 = np.log( intensity.I[ len( intensity.I ) //4 ] )
    ymax2 = Ymax + ( ymax - ymin ) *0.02

    ax2.set_xlim( xmin2, xmax2 )
    ax2.set_ylim( ymin2, ymax2)

    plt.tight_layout()
    plt.show( block=not auto )
    if auto:
        time.sleep( 0.5 )
        plt.close()
